@@ -3,14 +3,14 @@
 set -x
 
 if [ "x${ACLOCAL_DIR}" != "x" ]; then
-	ACLOCAL_ARG=-I ${ACLOCAL_DIR}
+	ACLOCAL_ARG="-I ${ACLOCAL_DIR}"
 fi
 
 ${ACLOCAL:-aclocal} ${ACLOCAL_ARG}
 
 # Darwin has different name for "libtoolize"
 UNAME=`uname`
-if [ "x${UNAME}" == "xDarwin" ]; then
+if [ "x${UNAME}" = "xDarwin" ]; then
 	${LIBTOOLIZE:-glibtoolize} -c --automake 
 else
 	${LIBTOOLIZE:-libtoolize} -c --automake 
