@@ -5,7 +5,7 @@
  *	Lu-chuan Kung and Kang-pen Chen.
  *	All rights reserved.
  *
- * Copyright (c) 2004, 2005, 2006
+ * Copyright (c) 2004, 2005, 2006, 2007, 2008
  *	libchewing Core Team. See ChangeLog for details.
  *
  * See the file "COPYING" for information on usage and redistribution
@@ -14,16 +14,20 @@
 
 #include <string.h>
 #include <sys/stat.h>
+/* ISO C99 Standard: 7.10/5.2.4.2.1 Sizes of integer types */
+#include <limits.h>
+#include <assert.h>
+#include <stdlib.h>
+#include <stdio.h>
 
 #include "chewing-utf8-util.h"
-#include "hash.h"
+#include "hash-private.h"
 #include "private.h"
 #include "global.h"
 
 int chewing_lifetime;
 
 static HASH_ITEM *hashtable[ HASH_TABLE_SIZE ];
-static char formatstring[ 30 ];
 static char hashfilename[ 200 ];
 
 int AlcUserPhraseSeq( UserPhraseData *pData, int phonelen, int wordlen )
@@ -36,6 +40,8 @@ int AlcUserPhraseSeq( UserPhraseData *pData, int phonelen, int wordlen )
 static int PhoneSeqTheSame( const uint16 p1[], const uint16 p2[] )
 {
 	int i;
+	if ( ! p1 || ! p2 )	/* FIXME: should not happend. */
+		return 0;
 
 	for ( i = 0; ( p1[ i ] != 0 && p2[ i ] != 0 ); i++ ) {
 		if ( p1[ i ] != p2[ i ] )
@@ -57,9 +63,9 @@ static unsigned int HashFunc( const uint16 phoneSeq[] )
 
 HASH_ITEM *HashFindPhonePhrase( const uint16 phoneSeq[], HASH_ITEM *pItemLast )
 {
-	HASH_ITEM *pNow = ( ! pItemLast ) ?
-		hashtable[ HashFunc( phoneSeq ) ] :
-		pItemLast->next;
+	HASH_ITEM *pNow = pItemLast ?
+			pItemLast->next :
+			hashtable[ HashFunc( phoneSeq ) ];
 	
 	for ( ; pNow; pNow = pNow->next ) 
 		if ( PhoneSeqTheSame( pNow->data.phoneSeq, phoneSeq ) )
@@ -113,6 +119,7 @@ HASH_ITEM *HashInsert( UserPhraseData *pData )
 	return pItem;
 }
 
+#ifdef ENABLE_DEBUG
 static void HashItem2String( char *str, HASH_ITEM *pItem )
 {
 	int i, len;
@@ -130,15 +137,16 @@ static void HashItem2String( char *str, HASH_ITEM *pItem )
 		pItem->data.maxfreq, pItem->data.origfreq );
 	strcat( str, buf );
 }
+#endif
 
 /* 
  * capacity of 'str' MUST bigger then FIELD_SIZE !
  */
-static void HashItem2Binary( char *str, HASH_ITEM *pItem )
+void HashItem2Binary( char *str, HASH_ITEM *pItem )
 {
 	int i, phraselen;
 	uint16 *pshort;
-	unsigned char buf[ FIELD_SIZE ], *puc;
+	unsigned char *puc;
 
 	memset(str, 0, FIELD_SIZE);
 	if ( sizeof(int) * 4 + ueStrLen( pItem->data.wordSeq ) * 2 +
@@ -181,9 +189,7 @@ void HashModify( HASH_ITEM *pItem )
 	fwrite( &chewing_lifetime, 1, 4, outfile );
 #ifdef ENABLE_DEBUG
 	sprintf( str, "%d", chewing_lifetime );
-	DEBUG_OUT( 
-		"HashModify-1: formatstring='%s',printing '%s'\n", 
-		formatstring, str );
+	DEBUG_OUT( "HashModify-1: '%-75s'\n", str );
 	DEBUG_FLUSH;
 #endif
 
@@ -195,14 +201,12 @@ void HashModify( HASH_ITEM *pItem )
 	}
 	else {
 		fseek( outfile,
-		       pItem->item_index * FIELD_SIZE + 4 + strlen( BIN_HASH_SIG ),
-		       SEEK_SET );
+			pItem->item_index * FIELD_SIZE + 4 + strlen( BIN_HASH_SIG ),
+			SEEK_SET );
 	}
 #ifdef ENABLE_DEBUG
 	HashItem2String( str, pItem );
-	DEBUG_OUT( 
-		"HashModify-2: formatstring='%s',printing '%s'\n",
-		formatstring, str );
+	DEBUG_OUT( "HashModify-2: '%-75s'\n", str );
 	DEBUG_FLUSH;
 #endif
 	HashItem2Binary( str, pItem );
@@ -234,56 +238,55 @@ static int isValidChineseString( char *str )
  */
 int ReadHashItem_bin( const char *srcbuf, HASH_ITEM *pItem, int item_index )
 {
-       int len, i, word_len, ptr;
-       uint16 *pshort;
-       char wordbuf[ 64 ];
-       unsigned char recbuf[ FIELD_SIZE ], *puc;
+	int len, i;
+	uint16 *pshort;
+	unsigned char recbuf[ FIELD_SIZE ], *puc;
 
-       memcpy( recbuf, srcbuf, FIELD_SIZE );
-       memset( pItem, 0, sizeof(HASH_ITEM) );
-       
-       /* freq info */
-       pItem->data.userfreq = *(int *) &recbuf[ 0 ];
-       pItem->data.recentTime = *(int *) &recbuf[ 4 ];
-       pItem->data.maxfreq = *(int *) &recbuf[ 8 ];
-       pItem->data.origfreq = *(int *) &recbuf[ 12 ];
+	memcpy( recbuf, srcbuf, FIELD_SIZE );
+	memset( pItem, 0, sizeof(HASH_ITEM) );
 
-       /* phone seq, length in num of chi words */
-       len = (int) recbuf[ 16 ];
-       pItem->data.phoneSeq = ALC( uint16, len + 1 );
-       pshort = (uint16 *) &recbuf[ 17 ];
-       for ( i = 0; i < len; i++ ) {
-               pItem->data.phoneSeq[ i ] = *pshort;
-               ++pshort;
-       }
-       pItem->data.phoneSeq[ i ] = 0;
-       
-       /* phrase, length in num of bytes */
-       puc = (unsigned char *) pshort;
-       pItem->data.wordSeq = ALC( char, (*puc) + 1 );
-       strcpy( pItem->data.wordSeq, (char *) (puc + 1) );
-       pItem->data.wordSeq[ (int) *puc ] = '\0';
+	/* freq info */
+	pItem->data.userfreq = *(int *) &recbuf[ 0 ];
+	pItem->data.recentTime = *(int *) &recbuf[ 4 ];
+	pItem->data.maxfreq = *(int *) &recbuf[ 8 ];
+	pItem->data.origfreq = *(int *) &recbuf[ 12 ];
 
-       /* Invalid UTF-8 Chinese characters found */
-       if ( ! isValidChineseString( pItem->data.wordSeq ) ) {
-               goto ignore_corrupted_record;
-       }
+	/* phone seq, length in num of chi words */
+	len = (int) recbuf[ 16 ];
+	pItem->data.phoneSeq = ALC( uint16, len + 1 );
+	pshort = (uint16 *) &recbuf[ 17 ];
+	for ( i = 0; i < len; i++ ) {
+		pItem->data.phoneSeq[ i ] = *pshort;
+		++pshort;
+	}
+	pItem->data.phoneSeq[ i ] = 0;
 
-       /* set item_index */
-       pItem->item_index = item_index;
+	/* phrase, length in num of bytes */
+	puc = (unsigned char *) pshort;
+	pItem->data.wordSeq = ALC( char, (*puc) + 1 );
+	strcpy( pItem->data.wordSeq, (char *) (puc + 1) );
+	pItem->data.wordSeq[ (int) *puc ] = '\0';
 
-       return 1; /* continue */
+	/* Invalid UTF-8 Chinese characters found */
+	if ( ! isValidChineseString( pItem->data.wordSeq ) ) {
+		goto ignore_corrupted_record;
+	}
+
+	/* set item_index */
+	pItem->item_index = item_index;
+
+	return 1; /* continue */
 
 ignore_corrupted_record:
-       if ( pItem->data.phoneSeq != NULL ) {
-               free( pItem->data.phoneSeq );
-               pItem->data.phoneSeq = NULL;
-       }
-       if ( pItem->data.wordSeq != NULL ) {
-               free( pItem->data.wordSeq );
-               pItem->data.wordSeq = NULL;
-       }
-       return -1; /* ignore */
+	if ( pItem->data.phoneSeq != NULL ) {
+		free( pItem->data.phoneSeq );
+		pItem->data.phoneSeq = NULL;
+	}
+	if ( pItem->data.wordSeq != NULL ) {
+		free( pItem->data.wordSeq );
+		pItem->data.wordSeq = NULL;
+	}
+	return -1; /* ignore */
 }
 
 /**
@@ -344,10 +347,10 @@ static FILE *open_file_get_length(
 		*size = ftell( tf );
 		fseek( tf, 0, SEEK_SET );
 	}
-	return  tf;
+	return tf;
 }
 
-static char *_load_hash_file( const char *filename, int *size )
+char *_load_hash_file( const char *filename, int *size )
 {
 	int flen;
 	char *pd = NULL;
@@ -381,8 +384,8 @@ static int migrate_hash_to_bin( const char *ofilename )
 {
 	FILE *txtfile;
 	char oldname[ 256 ], *dump, *seekdump;
-	HASH_ITEM item, *pItem;
-	int item_index, hashvalue, iret, tflen;
+	HASH_ITEM item;
+	int item_index, iret, tflen;
 
 	/* allocate dump buffer */
 	txtfile = open_file_get_length( ofilename, "r", &tflen );
@@ -436,10 +439,12 @@ static int migrate_hash_to_bin( const char *ofilename )
 	fwrite( dump, seekdump - dump, 1, txtfile );
 	fflush( txtfile );
 	fclose( txtfile );
+	free( dump );
 
 	return  1;
 }
 
+#if 0
 /**
  * Attempt to re-compute lifetime
  */
@@ -477,15 +482,32 @@ static int ComputeChewingLifeTime()
        }
        return 0;
 }
+#endif
 
-int ReadHash( const char *path )
+static HASH_ITEM *pHead = NULL;
+
+static void TerminateHash()
 {
-	HASH_ITEM item, *pItem;
-	int item_index, hashvalue, iret, fsize, hdrlen;
+	HASH_ITEM *pItem;
+	while ( pHead ) {
+		pItem = pHead;
+		pHead = pItem->next;
+		DEBUG_CHECKPOINT();
+		free( pItem->data.phoneSeq );
+		free( pItem->data.wordSeq );
+		free( pItem );
+	}
+	pHead = NULL;
+}
+
+int InitHash( const char *path )
+{
+	HASH_ITEM item, *pItem, *pPool = NULL;
+	int item_index, hashvalue, iret, fsize, hdrlen, oldest = INT_MAX;
 	char *dump, *seekdump;
 
 	/* make sure of write permission */
-	if ( access( path, W_OK ) != 0) {
+	if ( access( path, W_OK ) != 0 ) {
 		if ( getenv( "HOME" ) ) {
 			sprintf(
 				hashfilename, "%s%s", 
@@ -502,8 +524,7 @@ int ReadHash( const char *path )
 	} else {
 		sprintf( hashfilename, "%s" PLAT_SEPARATOR "%s", path, HASH_FILE );
 	}
-	memset( hashtable, 0, HASH_TABLE_SIZE );
-	sprintf( formatstring, "%%-%ds", FIELD_SIZE );
+	memset( hashtable, (int) NULL, HASH_TABLE_SIZE );
 
 open_hash_file:
 	dump = _load_hash_file( hashfilename, &fsize );
@@ -520,11 +541,12 @@ open_hash_file:
 		}
 		chewing_lifetime = 0;
 		fwrite( BIN_HASH_SIG, 1, strlen( BIN_HASH_SIG ), outfile );
-		fwrite( &chewing_lifetime, 1, sizeof(chewing_lifetime), outfile );
+		fwrite( &chewing_lifetime, 1,
+		                sizeof(chewing_lifetime), outfile );
 		fclose( outfile );
 	}
 	else {
-		char header[ 5 ];
+		static int is_set_pHead = 0;
 		if ( memcmp(dump, BIN_HASH_SIG, strlen(BIN_HASH_SIG)) != 0 ) {
 			/* perform migrate from text-based to binary form */
 			free( dump );
@@ -550,17 +572,36 @@ open_hash_file:
 			else if ( iret == 0 )
 				break;
 
-			hashvalue = HashFunc( item.data.phoneSeq );
 			pItem = ALC( HASH_ITEM, 1 );
 			memcpy( pItem, &item, sizeof( HASH_ITEM ) );
-			pItem->next = hashtable[ hashvalue ];
-			hashtable[ hashvalue ] = pItem;
+			pItem->next = pPool;
+			pPool = pItem;
+			if ( ! is_set_pHead ) {
+				pHead = pItem;
+				is_set_pHead = 1;
+			}
+
+			if ( oldest > pItem->data.recentTime ) {
+				oldest = pItem->data.recentTime;
+			}
+
 			seekdump += FIELD_SIZE;
 			fsize -= FIELD_SIZE;
 		}
 		free( dump );
-		ComputeChewingLifeTime();
+
+		while ( pPool ) {
+			pItem = pPool;
+			pPool = pItem->next;
+
+			hashvalue = HashFunc( pItem->data.phoneSeq );
+			pItem->next = hashtable[ hashvalue ];
+			hashtable[ hashvalue ] = pItem;
+			pItem->data.recentTime -= oldest;
+		}
+		chewing_lifetime -= oldest;
 	}
+	addTerminateService( TerminateHash );
 	return 1;
 }
 
